"""
This package contains different kinds of adapters.

json
    This package offers an adapter for serialization and deserialization of PyI40AAS objects to/from JSON.
xml
    This package offers an adapter for serialization and deserialization of PyI40AAS objects to/from XML.
"""
