# Copyright (c) 2019 PyI40AAS Contributors
#
# This program and the accompanying materials are made available under the terms of the Eclipse Public License v. 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0 which is available
# at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
"""
This module contains the security aspects of the AAS metamodel. Currently, the security model is not ready yet, so this
module doesn't do anything.
"""


class Security:
    """
    Security model is not ready yet. This is just a placeholder class.
    """
    pass
